(* Applies the conversion at all depths until saturated.
   Written since DEPTH_CONV doesn't work eg with:
   NUM_NORMALIZE_CONV `2 DIV (2 * (x * y + 2) + 7 * x * y) DIV 2` 
   Top - down version *)
let rec my_depth_conv conv term =
  let converted = try conv term with _ -> REFL term in
  let top = get_res converted in
  if top <> term then converted else
  try COMB_CONV (my_depth_conv conv) term
  with Failure _ -> 
    try ABS_CONV (my_depth_conv conv) term
    with Failure _ -> REFL term
;;

(* Bottom - up verion *)
let rec my_depth_conv_up conv term =
  let converted = 
    try COMB_CONV (my_depth_conv_up conv) term
    with Failure _ -> 
      try ABS_CONV (my_depth_conv_up conv) term
      with Failure _ -> REFL term
  in
  let top = get_res converted in
  if top <> term then converted else
  try conv term with _ -> REFL term
;;

(* Tries conversions from a list at all depths until saturated *)
let rec try_conv_list cl term =
  let fold_fun thm conv =
    try TRANS thm (conv (get_res thm))
    with _ -> thm
  in
  let thm = List.fold_left fold_fun (REFL term) cl in
  if get_res thm = term then thm else
  TRANS thm (try_conv_list cl (get_res thm))
;;

let cas_net = ref (empty_net);;

let cas_pattern pat conv =
  cas_net := net_of_conv pat conv !cas_net
;;


let cas_assume_pattern pat conv =
  let current_conv term =
    let assums = List.fold_left (fun l r -> mk_binop `(==>)` r l) term !assumptions in
    let thm = conv assums in
    let ret = REWRITE_RULE (List.map ASSUME !assumptions) thm in
    try ignore (get_res ret); ret 
    with Failure _ -> EQT_INTRO ret
  in
  cas_net := net_of_conv pat current_conv !cas_net
;;


let cas_old_conv term = 
  let conv term =
    let lst = lookup term !cas_net in
    try_conv_list (List.map snd lst) term
  in
  try_conv_list [my_depth_conv conv] term
;;

(* Fast version, failing if unable to convert *)

let numeral_term = `NUMERAL`;;

let rec cas_cache_conv verbose term =
(*  print_qterm term; print_newline ();*)
  try 
    match Hashtbl.find cas_cache term with
    | Some thm -> thm
    | None -> failwith "Not simplifiable"
  with Not_found ->
    let test_pair (l, r) =
      let lthm = try cas_cache_conv verbose l with Failure "Not simplifiable" -> REFL l
      and rthm = try cas_cache_conv verbose r with Failure "Not simplifiable" -> REFL r in
      let thm = REWRITE_CONV [lthm; rthm] term in
      if get_res thm = term then begin
        failwith "Unable to simplify any side" 
      end else begin
        let rep_thm = 
          try TRANS thm (cas_cache_conv verbose (get_res thm))
          with Failure "Not simplifiable" -> thm
        in
        (*print_thm rep_thm; print_newline ();*)
        Hashtbl.replace cas_cache term (Some rep_thm);
        rep_thm
      end
    in
    try 
      let l, r = dest_comb term in
      if l = numeral_term then failwith "Unable to simplify nums" else
      let str = try fst (dest_const l) with _ -> "" in
      if str = "assuming" then failwith "Not descending inside assuming" else
      test_pair (l, r)
    with Failure x -> (* Not a combination or it didn't simplify *)
    try test_pair (dest_abs term)
    with Failure x -> (* Not a combination nor abstraction or they didn't simplify *)
      let lst = lookup term !cas_net in
      let fold_fun (changed, thm) conv =
        if changed then (true, thm) else
        try 
          let new_thm = conv term in
          let lt, rt = dest_eq (concl new_thm) in
          if ((rt = term)) then (false, thm)
          else if ((lt <> term)) then (false, thm)
          else (true, new_thm)
        with _ -> (false, thm)
      in
      let (changed, thm) = List.fold_left fold_fun (false, REFL term) (List.map snd lst) in
      if changed = false then begin
        Hashtbl.replace cas_cache term None;
        failwith "Not simplifiable" 
      end else begin
        let rep_thm =
          try TRANS thm (cas_cache_conv verbose (get_res thm))
          with Failure "Not simplifiable" -> thm
        in
        if verbose then (print_thm rep_thm; print_newline ()) else ();
        Hashtbl.replace cas_cache term (Some rep_thm);
        rep_thm 
      end
;;

let cas_conv term =
  try 
    let ret = cas_cache_conv false term in
(*    List.iter print_qterm !assumptions; print_thm ret; print_newline (); *)
    ret
  with Failure "Not simplifiable" -> 
(*    List.iter print_qterm !assumptions; print_qterm term; print_newline (); *)
    REFL term
;;

(* TODO: This is a hack, try to fit ETA inside the schema *)
let cas_conv term =
  type_invention_warning := false;
  let ret = cas_conv term in
  let t2 = REWRITE_CONV [ETA_AX] (get_res ret) in
  let t3 = cas_conv (get_res t2) in
  let ret = TRANS ret (TRANS t2 t3) in
  type_invention_warning := true;
  ret
;;

let cas_eq_thm thm =
  let subthms = CONJUNCTS (SPEC_ALL thm) in
  let adder thm =
    let thm = SPEC_ALL thm in
    let term = concl thm in
    let (lo, rhs) = dest_comb term in
    let (op, lhs) = dest_comb lo in
    let current_conv term =
      let ret = REWRITE_CONV [thm] term in
      ret
    in
    cas_pattern lhs current_conv
  in
  List.iter adder subthms
;;


let impl_conv thm term =
  let thm = PART_MATCH (fst o dest_eq o snd o dest_imp) thm term in
  let needed = fst (dest_imp (concl thm)) in
  let needed_thm = cas_cache_conv false needed in
  if get_res (needed_thm) <> `T` then failwith "impl_conv fails" else
  MP thm (EQT_ELIM needed_thm)
;;



let cas_impl_thm thm =  
  let subthms = CONJUNCTS (SPEC_ALL thm) in
  let adder thm =
    let thm = SPEC_ALL thm in
    let term = snd (dest_imp (concl thm)) in
    let (lo, rhs) = dest_comb term in
    let (op, lhs) = dest_comb lo in
    let current_conv term =
      impl_conv thm term
    in
    cas_pattern lhs current_conv
  in
  List.iter adder subthms
;;



(*
let cas_print_net () =
  let rec print = function
    Netnode ([], []) -> ()
  | Netnode (h::t, []) -> match h with
    
  in print !cas_net
;;
  *)
